@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2014-2022 Savoir-faire Linux Inc.
+ *  Copyright (C) 2014-2023 Savoir-faire Linux Inc.
  *  Author: Sébastien Blin <sebastien.blin@savoirfairelinux.com>
  *          Adrien Béraud <adrien.beraud@savoirfairelinux.com>
  *          Vsevolod Ivanov <vsevolod.ivanov@savoirfairelinux.com>
@@ -23,10 +23,9 @@
 #include "op_cache.h"
 #include "utils.h"
 
-#include <http_parser.h>
+#include <llhttp.h>
 #include <deque>
 
-
 namespace dht {
 
 struct DhtProxyClient::InfoState {
@@ -98,7 +97,7 @@ std::string
 getRandomSessionId(size_t length = 8) {
     static constexpr const char chars[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789!#$%&()*+,./:;<=>?@[]^_`{|}~";
     std::string str(length, 0);
-    crypto::random_device rdev;
+    std::random_device rdev;
     std::uniform_int_distribution<> dist(0, (sizeof(chars)/sizeof(char)) - 2);
     std::generate_n( str.begin(), length, [&]{ return chars[dist(rdev)]; } );
     return str;
@@ -107,13 +106,26 @@ getRandomSessionId(size_t length = 8) {
 DhtProxyClient::DhtProxyClient() {}
 
 DhtProxyClient::DhtProxyClient(
-        std::shared_ptr<dht::crypto::Certificate> serverCA, dht::crypto::Identity clientIdentity,
-        std::function<void()> signal, const std::string& serverHost,
-        const std::string& pushClientId, std::shared_ptr<dht::Logger> logger)
+        std::shared_ptr<dht::crypto::Certificate> serverCA,
+        dht::crypto::Identity clientIdentity,
+        std::function<void()> signal,
+        const std::string& serverHost,
+        const std::string& userAgent,
+        const std::string& pushClientId,
+        const std::string& pushToken,
+        const std::string& pushTopic,
+        const std::string& pushPlatform,
+        std::shared_ptr<dht::Logger> logger
+)
     : DhtInterface(logger)
     , proxyUrl_(serverHost)
     , clientIdentity_(clientIdentity), serverCertificate_(serverCA)
-    , pushClientId_(pushClientId), pushSessionId_(getRandomSessionId())
+    , userAgent_(userAgent)
+    , pushClientId_(pushClientId)
+    , pushSessionId_(getRandomSessionId())
+    , deviceKey_(pushToken)
+    , notificationTopic_(pushTopic)
+    , platform_(pushPlatform)
     , loopSignal_(signal)
     , jsonReader_(Json::CharReaderBuilder{}.newCharReader())
 {
@@ -191,10 +203,12 @@ void
 DhtProxyClient::stop()
 {
     if (not isDestroying_.exchange(true)) {
+        std::shared_ptr<http::Resolver> resolver;
         {
             std::lock_guard<std::mutex> l(resolverLock_);
-            resolver_.reset();
+            resolver = std::move(resolver_);
         }
+        resolver->cancel();
         cancelAllListeners();
         if (infoState_)
             infoState_->cancel = true;
@@ -483,7 +497,7 @@ DhtProxyClient::buildRequest(const std::string& target)
         request->set_certificate_authority(serverCertificate_);
     if (clientIdentity_.first and clientIdentity_.second)
         request->set_identity(clientIdentity_);
-    request->set_header_field(restinio::http_field_t::user_agent, "RESTinio client");
+    request->set_header_field(restinio::http_field_t::user_agent, userAgent_);
     return request;
 }
 
@@ -491,7 +505,7 @@ void
 DhtProxyClient::doPut(const InfoHash& key, Sp<Value> val, DoneCallbackSimple cb, time_point /*created*/, bool permanent)
 {
     if (logger_)
-        logger_->d("[proxy:client] [put] [search %s] executing for %s", key.to_c_str(), val->toString().c_str());
+        logger_->debug("[proxy:client] [put] [search {}] executing for {}", key, val->toString());
 
     try {
         auto request = buildRequest("/key/" + key.toString());
@@ -539,12 +553,12 @@ DhtProxyClient::doPut(const InfoHash& key, Sp<Value> val, DoneCallbackSimple cb,
                         }
                     } else {
                         if (logger_)
-                            logger_->e("[proxy:client] [status] failed to parse value from  server", response.status_code);
+                            logger_->error("[proxy:client] [put] failed to parse value from server: {}", err);
                     }
                 }
             } else {
                 if (logger_)
-                    logger_->e("[proxy:client] [status] failed with code=%i", response.status_code);
+                    logger_->error("[proxy:client] [put] failed with code={:d}", response.status_code);
                 if (not response.aborted and response.status_code == 0)
                     opFailed();
             }
@@ -563,7 +577,8 @@ DhtProxyClient::doPut(const InfoHash& key, Sp<Value> val, DoneCallbackSimple cb,
     }
     catch (const std::exception &e){
         if (logger_)
-            logger_->e("[proxy:client] [put %s] error: %s", key.to_c_str(), e.what());
+            logger_->error("[proxy:client] [put {}] error: {}", key, e.what());
+        opFailed();
     }
 }
 
@@ -721,7 +736,7 @@ DhtProxyClient::onProxyInfos(const Json::Value& proxyInfos, const sa_family_t fa
         }
     } else {
         if (logger_)
-            logger_->d("[proxy:client] [info] got proxy reply for %s",
+            logger_->debug("[proxy:client] [info] got proxy reply for {}",
                        family == AF_INET ? "ipv4" : "ipv6");
         try {
             myid = InfoHash(proxyInfos["node_id"].asString());
@@ -729,6 +744,7 @@ DhtProxyClient::onProxyInfos(const Json::Value& proxyInfos, const sa_family_t fa
             stats6_ = NodeStats(proxyInfos["ipv6"]);
             auto publicIp = parsePublicAddress(proxyInfos["public_ip"]);
             ipChanged = pubAddress && pubAddress.toString() != publicIp.toString();
+            bool pub = ipChanged || !pubAddress;
             pubAddress = publicIp;
 
             if (proxyInfos.isMember("local_ip")) {
@@ -739,6 +755,18 @@ DhtProxyClient::onProxyInfos(const Json::Value& proxyInfos, const sa_family_t fa
                 }
             }
 
+            if (pub && publicAddressChangedCb_) {
+                std::vector<SockAddr> addresses;
+                if (publicAddressV4_)
+                    addresses.emplace_back(publicAddressV4_);
+                if (publicAddressV6_)
+                    addresses.emplace_back(publicAddressV6_);
+                std::lock_guard<std::mutex> lock(lockCallbacks_);
+                callbacks_.emplace_back([cb=publicAddressChangedCb_, addresses = std::move(addresses)](){
+                    cb(std::move(addresses));
+                });
+            }
+
             if (!ipChanged && stats4_.good_nodes + stats6_.good_nodes)
                 status = NodeStatus::Connected;
             else if (!ipChanged && stats4_.dubious_nodes + stats6_.dubious_nodes)
@@ -785,8 +813,8 @@ SockAddr
 DhtProxyClient::parsePublicAddress(const Json::Value& val)
 {
     auto public_ip = val.asString();
-    auto hostAndService = splitPort(public_ip);
-    auto sa = SockAddr::resolve(hostAndService.first);
+    auto [host, service] = splitPort(public_ip);
+    auto sa = SockAddr::resolve(host);
     if (sa.empty()) return {};
     return sa.front().getMappedIPv4();
 }
@@ -995,7 +1023,10 @@ DhtProxyClient::handleExpireListener(const asio::error_code &ec, const InfoHash&
             }
         } else {
             // stop the request
-            listener.request.reset();
+            if (listener.request) {
+                listener.request->cancel();
+                listener.request.reset();
+            }
         }
         search->second.listeners.erase(it);
         if (logger_)
@@ -1090,6 +1121,12 @@ DhtProxyClient::sendListen(const restinio::http_request_header_t& header,
             std::lock_guard<std::mutex> l(requestLock_);
             requests_[reqid] = request;
         }
+        request->add_on_status_callback([request, seconds = this->listenKeepIdle()] (unsigned status_code) {
+            if(status_code == 200) {
+                // increase TCP_KEEPIDLE to save power
+                request->get_connection()->set_keepalive(seconds);
+            }
+        });
         request->send();
     }
     catch (const std::exception &e){
@@ -1141,7 +1178,7 @@ DhtProxyClient::restartListeners(const asio::error_code &ec)
 
     std::lock_guard<std::mutex> lock(searchLock_);
     for (auto& search : searches_) {
-        auto key = search.first;
+        const auto& key = search.first;
         for (auto& put : search.second.puts) {
             doPut(key, put.second.value, [ok = put.second.ok](bool result){
                 *ok = result;
@@ -1153,6 +1190,10 @@ DhtProxyClient::restartListeners(const asio::error_code &ec)
             put.second.refreshPutTimer->async_wait(std::bind(&DhtProxyClient::handleRefreshPut, this,
                                                    std::placeholders::_1, key, put.first));
         }
+        // Restart failed pending puts
+        for (auto& pput : search.second.pendingPuts) {
+            doPut(key, pput, {}, time_point::max(), true);
+        }
     }
     if (not deviceKey_.empty()) {
         if (logger_)
@@ -1192,9 +1233,10 @@ DhtProxyClient::restartListeners(const asio::error_code &ec)
     }
 }
 
-void
-DhtProxyClient::pushNotificationReceived(const std::map<std::string, std::string>& notification)
+PushNotificationResult
+DhtProxyClient::pushNotificationReceived([[maybe_unused]] const std::map<std::string, std::string>& notification)
 {
+    auto ret = PushNotificationResult::IgnoredNoOp;
 #ifdef OPENDHT_PUSH_NOTIFICATIONS
     {
         // If a push notification is received, the proxy is up and running
@@ -1211,7 +1253,7 @@ DhtProxyClient::pushNotificationReceived(const std::map<std::string, std::string
         if (sessionId != notification.end() and sessionId->second != pushSessionId_) {
             if (logger_)
                 logger_->d("[proxy:client] [push] ignoring push for other session");
-            return;
+            return PushNotificationResult::IgnoredWrongSession;
         }
         std::lock_guard<std::mutex> lock(searchLock_);
         auto timeout = notification.find("timeout");
@@ -1228,10 +1270,12 @@ DhtProxyClient::pushNotificationReceived(const std::map<std::string, std::string
                 else
                     put.refreshPutTimer->expires_at(std::chrono::steady_clock::now());
                 put.refreshPutTimer->async_wait(std::bind(&DhtProxyClient::handleRefreshPut, this, std::placeholders::_1, key, vid));
+                ret = PushNotificationResult::PutRefresh;
             } else {
                 // Refresh listen
                 for (auto& list : search.listeners)
                     resubscribe(key, list.first, list.second);
+                ret = search.listeners.empty() ? PushNotificationResult::IgnoredNoOp : PushNotificationResult::ListenRefresh;
             }
         } else {
             auto key = InfoHash(notification.at("key"));
@@ -1240,12 +1284,12 @@ DhtProxyClient::pushNotificationReceived(const std::map<std::string, std::string
                 sendTime = system_clock::time_point(std::chrono::milliseconds(std::stoull(notification.at("t"))));
             } catch (...) {}
             auto& search = searches_.at(key);
+            auto expired = notification.find("exp");
             for (auto& list : search.listeners) {
                 if (list.second.opstate->stop)
                     continue;
                 if (logger_)
                     logger_->d("[proxy:client] [push] [search %s] received", key.to_c_str());
-                auto expired = notification.find("exp");
                 auto token = list.first;
                 auto opstate = list.second.opstate;
                 if (expired == notification.end()) {
@@ -1258,6 +1302,7 @@ DhtProxyClient::pushNotificationReceived(const std::map<std::string, std::string
                         // present in the new list
                         cb(oldValues, true, sendTime);
                     });
+                    ret = PushNotificationResult::Values;
                 } else {
                     std::stringstream ss(expired->second);
                     std::vector<Value::Id> ids;
@@ -1283,22 +1328,25 @@ DhtProxyClient::pushNotificationReceived(const std::map<std::string, std::string
                         });
                     }
                     launchLoop = true;
+                    ret = PushNotificationResult::ValuesExpired;
                 }
             }
         }
+    } catch (const std::out_of_range& e) {
+        ret = PushNotificationResult::IgnoredNoOp;
     } catch (const std::exception& e) {
         if (logger_)
             logger_->e("[proxy:client] [push] receive error: %s", e.what());
+        ret = PushNotificationResult::Error;
     }
     if (launchLoop)
         loopSignal_();
-#else
-    (void) notification;
 #endif
+    return ret;
 }
 
 void
-DhtProxyClient::resubscribe(const InfoHash& key, const size_t token, Listener& listener)
+DhtProxyClient::resubscribe([[maybe_unused]] const InfoHash& key, [[maybe_unused]] const size_t token, [[maybe_unused]] Listener& listener)
 {
 #ifdef OPENDHT_PUSH_NOTIFICATIONS
     if (deviceKey_.empty())
@@ -1322,9 +1370,6 @@ DhtProxyClient::resubscribe(const InfoHash& key, const size_t token, Listener& l
                                                 std::placeholders::_1, key, token, opstate));
     auto vcb = listener.cb;
     sendListen(header, vcb, opstate, listener, ListenMethod::RESUBSCRIBE);
-#else
-    (void) key;
-    (void) listener;
 #endif
 }
 
@@ -1358,4 +1403,24 @@ DhtProxyClient::fillBody(bool resubscribe)
 }
 #endif // OPENDHT_PUSH_NOTIFICATIONS
 
+void
+DhtProxyClient::setPushNotificationToken([[maybe_unused]] const std::string& token) {
+#ifdef OPENDHT_PUSH_NOTIFICATIONS
+    std::unique_lock<std::mutex> l(lockCurrentProxyInfos_);
+    if (deviceKey_ != token) {
+        deviceKey_ = token;
+        if (statusIpv4_ == NodeStatus::Connected || statusIpv6_ == NodeStatus::Connected) {
+            if (logger_)
+                logger_->d("[proxy:client] [push] token changed, resubscribing");
+            for (auto& search : searches_) {
+                for (auto& listener : search.second.listeners) {
+                    resubscribe(search.first, listener.first, listener.second);
+                }
+            }
+        }
+    }
+#endif
+}
+
+
 } // namespace dht
