@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2014-2020 Savoir-faire Linux Inc.
+ *  Copyright (C) 2014-2023 Savoir-faire Linux Inc.
  *  Author(s) : Adrien Béraud <adrien.beraud@savoirfairelinux.com>
  *              Simon Désaulniers <simon.desaulniers@savoirfairelinux.com>
  *
@@ -20,10 +20,12 @@
 #include "network_engine.h"
 #include "request.h"
 #include "default_types.h"
-#include "log_enable.h"
+#include "logger.h"
 #include "parsed_message.h"
 
 #include <msgpack.hpp>
+#include <chrono>
+#include <string_view>
 
 namespace dht {
 namespace net {
@@ -41,7 +43,8 @@ constexpr std::chrono::seconds NetworkEngine::UDP_REPLY_TIME;
 constexpr std::chrono::seconds NetworkEngine::RX_MAX_PACKET_TIME;
 constexpr std::chrono::seconds NetworkEngine::RX_TIMEOUT;
 
-const std::string NetworkEngine::my_v {"RNG1"};
+/* OpenDHT User Agent (UA) */
+constexpr std::string_view OPENDHT_UA {"o2"};
 
 static constexpr uint8_t v4prefix[16] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xFF, 0xFF, 0, 0, 0, 0};
 
@@ -116,7 +119,7 @@ NetworkEngine::~NetworkEngine() {
 }
 
 void
-NetworkEngine::tellListener(Sp<Node> node, Tid socket_id, const InfoHash& hash, want_t want,
+NetworkEngine::tellListener(const Sp<Node>& node, Tid socket_id, const InfoHash& hash, want_t want,
         const Blob& ntoken, std::vector<Sp<Node>>&& nodes,
         std::vector<Sp<Node>>&& nodes6, std::vector<Sp<Value>>&& values,
         const Query& query, int version)
@@ -124,9 +127,9 @@ NetworkEngine::tellListener(Sp<Node> node, Tid socket_id, const InfoHash& hash,
     auto nnodes = bufferNodes(node->getFamily(), hash, want, nodes, nodes6);
     try {
         if (version >= 1) {
-            sendUpdateValues(node, hash, values, scheduler.time(), ntoken, socket_id);
+            sendUpdateValues(node, hash, std::move(values), scheduler.time(), ntoken, socket_id);
         } else {
-            sendNodesValues(node->getAddr(), socket_id, nnodes.first, nnodes.second, values, query, ntoken);
+            sendNodesValues(node->getAddr(), socket_id, nnodes.first, nnodes.second, std::move(values), query, ntoken);
         }
     } catch (const std::overflow_error& e) {
         if (logger_)
@@ -135,12 +138,12 @@ NetworkEngine::tellListener(Sp<Node> node, Tid socket_id, const InfoHash& hash,
 }
 
 void
-NetworkEngine::tellListenerRefreshed(Sp<Node> n, Tid socket_id, const InfoHash&, const Blob& token, const std::vector<Value::Id>& values, int version)
+NetworkEngine::tellListenerRefreshed(const Sp<Node>& n, Tid socket_id, const InfoHash&, const Blob& token, const std::vector<Value::Id>& values, int version)
 {
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
 
-    pk.pack_map(4 + (version >= 1 ? 1 : 0) + (config.network?1:0));
+    pk.pack_map(4 + (version >= 1 ? 1 : 0) + (config.network?1:0)+(config.is_client?1:0));
 
     pk.pack(version >= 1 ? KEY_A : KEY_U);
         pk.pack_map(1 + (version >= 1 ? 1 : 0) + (not values.empty()?1:0) + (not token.empty()?1:0));
@@ -159,10 +162,13 @@ NetworkEngine::tellListenerRefreshed(Sp<Node> n, Tid socket_id, const InfoHash&,
         }
 
     pk.pack(KEY_Y); pk.pack(version >= 1 ? KEY_Q : KEY_R);
-    pk.pack(KEY_UA); pk.pack(my_v);
+    pk.pack(KEY_UA); pk.pack(OPENDHT_UA);
     if (config.network) {
         pk.pack(KEY_NETID); pk.pack(config.network);
     }
+    if (config.is_client) {
+        pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+    }
 
     if (version >= 1) {
         Tid tid (n->getNewTid());
@@ -186,12 +192,12 @@ NetworkEngine::tellListenerRefreshed(Sp<Node> n, Tid socket_id, const InfoHash&,
 }
 
 void
-NetworkEngine::tellListenerExpired(Sp<Node> n, Tid socket_id, const InfoHash&, const Blob& token, const std::vector<Value::Id>& values, int version)
+NetworkEngine::tellListenerExpired(const Sp<Node>& n, Tid socket_id, const InfoHash&, const Blob& token, const std::vector<Value::Id>& values, int version)
 {
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
 
-    pk.pack_map(4 + (version >= 1 ? 1 : 0) + (config.network?1:0));
+    pk.pack_map(4 + (version >= 1 ? 1 : 0) + (config.network?1:0)+(config.is_client?1:0));
 
     pk.pack(version >= 1 ? KEY_A : KEY_U);
         pk.pack_map(1 + (version >= 1 ? 1 : 0) + (not values.empty()?1:0) + (not token.empty()?1:0));
@@ -210,16 +216,19 @@ NetworkEngine::tellListenerExpired(Sp<Node> n, Tid socket_id, const InfoHash&, c
         }
 
     pk.pack(KEY_Y); pk.pack(version >= 1 ? KEY_Q : KEY_R);
-    pk.pack(KEY_UA); pk.pack(my_v);
+    pk.pack(KEY_UA); pk.pack(OPENDHT_UA);
     if (config.network) {
         pk.pack(KEY_NETID); pk.pack(config.network);
     }
+    if (config.is_client) {
+        pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+    }
 
     if (version >= 1) {
         Tid tid (n->getNewTid());
 
         pk.pack(KEY_Q);   pk.pack(QUERY_UPDATE);
-        pk.pack(KEY_TID);     pk.pack(tid);
+        pk.pack(KEY_TID); pk.pack(tid);
 
         auto req = std::make_shared<Request>(MessageType::UpdateValue, tid, n,
             Blob(buffer.data(), buffer.data() + buffer.size()),
@@ -375,7 +384,6 @@ NetworkEngine::isMartian(const SockAddr& addr)
         const auto& sin6 = addr.getIPv6();
         const uint8_t* address = (const uint8_t*)&sin6.sin6_addr;
         return address[0] == 0xFF ||
-              (address[0] == 0xFE && (address[1] & 0xC0) == 0x80) ||
                memcmp(address, InfoHash::zero().data(), 16) == 0 ||
                memcmp(address, v4prefix,      12) == 0;
     }
@@ -415,7 +423,7 @@ NetworkEngine::processMessage(const uint8_t *buf, size_t buflen, SockAddr f)
         return;
     }
 
-    std::unique_ptr<ParsedMessage> msg {new ParsedMessage};
+    auto msg = std::make_unique<ParsedMessage>();
     try {
         msgpack::unpacked msg_res = msgpack::unpack((const char*)buf, buflen);
         msg->msgpack_unpack(msg_res.get());
@@ -460,9 +468,9 @@ NetworkEngine::processMessage(const uint8_t *buf, size_t buflen, SockAddr f)
                     // process the full message
                     process(std::move(pmsg_it->second.msg), from);
                     partial_messages.erase(pmsg_it);
-		} catch (...) {
+                } catch (...) {
                     return;
-		}  
+                }
             } else
                 scheduler.add(now + RX_TIMEOUT, std::bind(&NetworkEngine::maintainRxBuffer, this, msg->tid));
         }
@@ -485,11 +493,11 @@ NetworkEngine::processMessage(const uint8_t *buf, size_t buflen, SockAddr f)
     }
 
     if (msg->value_parts.empty()) {
-        try {	    
+        try {
             process(std::move(msg), from);
-	} catch(...) {
+        } catch(...) {
             return;
-	}
+        }
     } else {
         // starting partial message session
         auto k = msg->tid;
@@ -501,9 +509,8 @@ NetworkEngine::processMessage(const uint8_t *buf, size_t buflen, SockAddr f)
             pmsg.last_part = now;
             scheduler.add(now + RX_MAX_PACKET_TIME, std::bind(&NetworkEngine::maintainRxBuffer, this, k));
             scheduler.add(now + RX_TIMEOUT, std::bind(&NetworkEngine::maintainRxBuffer, this, k));
-        } else
-            if (logger_)
-                logger_->e("Partial message with given TID already exists");
+        } else if (logger_)
+            logger_->e("Partial message with given TID %u already exists", k);
     }
 }
 
@@ -519,8 +526,13 @@ NetworkEngine::process(std::unique_ptr<ParsedMessage>&& msg, const SockAddr& fro
             throw DhtProtocolException {DhtProtocolException::UNKNOWN_TID, "Can't find socket", msg->id};
         node->received(now, {});
         onNewNode(node, 2);
-        deserializeNodes(*msg, from);
-        rsocket->on_receive(node, std::move(*msg));
+        try {
+            deserializeNodes(*msg, from);
+            rsocket->on_receive(node, std::move(*msg));
+        } catch (const DhtProtocolException& e) {
+            if (logger_)
+                logger_->w("Can't deserialize nodes %s", e.what());
+        }
     }
     else if (msg->type == MessageType::Error or msg->type == MessageType::Reply) {
         auto rsocket = node->getSocket(msg->tid);
@@ -583,13 +595,22 @@ NetworkEngine::process(std::unique_ptr<ParsedMessage>&& msg, const SockAddr& fro
                     r.node->authSuccess();
                 }
                 r.reply_time = scheduler.time();
-
-                deserializeNodes(*msg, from);
-                r.setDone(std::move(*msg));
+                try {
+                    deserializeNodes(*msg, from);
+                    r.setDone(std::move(*msg));
+                } catch (const DhtProtocolException& e) {
+                    if (logger_)
+                        logger_->w("Can't deserialize nodes %s", e.what());
+                }
                 break;
             } else { /* request socket data */
-                deserializeNodes(*msg, from);
-                rsocket->on_receive(node, std::move(*msg));
+                try {
+                    deserializeNodes(*msg, from);
+                    rsocket->on_receive(node, std::move(*msg));
+                } catch (const DhtProtocolException& e) {
+                    if (logger_)
+                        logger_->w("Can't deserialize nodes %s", e.what());
+                }
             }
             break;
         default:
@@ -699,11 +720,11 @@ NetworkEngine::send(const SockAddr& addr, const char *buf, size_t len, bool conf
 }
 
 Sp<Request>
-NetworkEngine::sendPing(Sp<Node> node, RequestCb&& on_done, RequestExpiredCb&& on_expired) {
+NetworkEngine::sendPing(const Sp<Node>& node, RequestCb&& on_done, RequestExpiredCb&& on_expired) {
     Tid tid (node->getNewTid());
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
-    pk.pack_map(5+(config.network?1:0));
+    pk.pack_map(5+(config.network?1:0)+(config.is_client?1:0));
 
     pk.pack(KEY_A); pk.pack_map(1);
      pk.pack(KEY_REQ_ID); pk.pack(myid);
@@ -711,10 +732,13 @@ NetworkEngine::sendPing(Sp<Node> node, RequestCb&& on_done, RequestExpiredCb&& o
     pk.pack(KEY_Q); pk.pack(QUERY_PING);
     pk.pack(KEY_TID); pk.pack(tid);
     pk.pack(KEY_Y); pk.pack(KEY_Q);
-    pk.pack(KEY_UA); pk.pack(my_v);
+    pk.pack(KEY_UA); pk.pack(OPENDHT_UA);
     if (config.network) {
         pk.pack(KEY_NETID); pk.pack(config.network);
     }
+    if (config.is_client) {
+        pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+    }
 
     auto req = std::make_shared<Request>(MessageType::Ping, tid, node,
         Blob(buffer.data(), buffer.data() + buffer.size()),
@@ -740,7 +764,7 @@ void
 NetworkEngine::sendPong(const SockAddr& addr, Tid tid) {
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
-    pk.pack_map(4+(config.network?1:0));
+    pk.pack_map(4+(config.network?1:0)+(config.is_client?1:0));
 
     pk.pack(KEY_R); pk.pack_map(2);
       pk.pack(KEY_REQ_ID); pk.pack(myid);
@@ -748,21 +772,24 @@ NetworkEngine::sendPong(const SockAddr& addr, Tid tid) {
 
     pk.pack(KEY_TID); pk.pack(tid);
     pk.pack(KEY_Y); pk.pack(KEY_R);
-    pk.pack(KEY_UA); pk.pack(my_v);
+    pk.pack(KEY_UA); pk.pack(OPENDHT_UA);
     if (config.network) {
         pk.pack(KEY_NETID); pk.pack(config.network);
     }
+    if (config.is_client) {
+        pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+    }
 
     send(addr, buffer.data(), buffer.size());
 }
 
 Sp<Request>
-NetworkEngine::sendFindNode(Sp<Node> n, const InfoHash& target, want_t want,
+NetworkEngine::sendFindNode(const Sp<Node>& n, const InfoHash& target, want_t want,
         RequestCb&& on_done, RequestExpiredCb&& on_expired) {
     Tid tid (n->getNewTid());
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
-    pk.pack_map(5+(config.network?1:0));
+    pk.pack_map(5+(config.network?1:0)+(config.is_client?1:0));
 
     pk.pack(KEY_A); pk.pack_map(2 + (want>0?1:0));
       pk.pack(KEY_REQ_ID);     pk.pack(myid);
@@ -777,10 +804,13 @@ NetworkEngine::sendFindNode(Sp<Node> n, const InfoHash& target, want_t want,
     pk.pack(KEY_Q); pk.pack(QUERY_FIND);
     pk.pack(KEY_TID); pk.pack(tid);
     pk.pack(KEY_Y); pk.pack(KEY_Q);
-    pk.pack(KEY_UA); pk.pack(my_v);
+    pk.pack(KEY_UA); pk.pack(OPENDHT_UA);
     if (config.network) {
         pk.pack(KEY_NETID); pk.pack(config.network);
     }
+    if (config.is_client) {
+        pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+    }
 
     auto req = std::make_shared<Request>(MessageType::FindNode, tid, n,
         Blob(buffer.data(), buffer.data() + buffer.size()),
@@ -802,12 +832,12 @@ NetworkEngine::sendFindNode(Sp<Node> n, const InfoHash& target, want_t want,
 
 
 Sp<Request>
-NetworkEngine::sendGetValues(Sp<Node> n, const InfoHash& info_hash, const Query& query, want_t want,
+NetworkEngine::sendGetValues(const Sp<Node>& n, const InfoHash& info_hash, const Query& query, want_t want,
         RequestCb&& on_done, RequestExpiredCb&& on_expired) {
     Tid tid (n->getNewTid());
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
-    pk.pack_map(5+(config.network?1:0));
+    pk.pack_map(5+(config.network?1:0)+(config.is_client?1:0));
 
     unsigned sendQuery = (not query.where.empty() or not query.select.empty()) ? 1 : 0;
     unsigned sendWant = (want > 0) ? 1 : 0;
@@ -830,11 +860,14 @@ NetworkEngine::sendGetValues(Sp<Node> n, const InfoHash& info_hash, const Query&
     pk.pack(KEY_Q); pk.pack(QUERY_GET);
     pk.pack(KEY_TID); pk.pack(tid);
     pk.pack(KEY_Y); pk.pack(KEY_Q);
-    pk.pack(KEY_UA); pk.pack(my_v);
+    pk.pack(KEY_UA); pk.pack(OPENDHT_UA);
     if (config.network) {
         pk.pack(KEY_NETID); pk.pack(config.network);
     }
-
+    if (config.is_client) {
+        pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+    }
+    
     auto req = std::make_shared<Request>(MessageType::GetValues, tid, n,
         Blob(buffer.data(), buffer.data() + buffer.size()),
         [=](const Request& req_status, ParsedMessage&& msg) { /* on done */
@@ -912,12 +945,16 @@ NetworkEngine::deserializeNodes(ParsedMessage& msg, const SockAddr& from) {
 }
 
 std::vector<Blob>
-NetworkEngine::packValueHeader(msgpack::sbuffer& buffer, const std::vector<Sp<Value>>& st)
+NetworkEngine::packValueHeader(msgpack::sbuffer& buffer, std::vector<Sp<Value>>::const_iterator b, std::vector<Sp<Value>>::const_iterator e) const
 {
-    auto svals = serializeValues(st);
+    std::vector<Blob> svals;
     size_t total_size = 0;
-    for (const auto& v : svals)
-        total_size += v.size();
+
+    svals.reserve(std::distance(b, e));
+    for (; b != e; ++b) {
+        svals.emplace_back(packMsg(*b));
+        total_size += svals.back().size();
+    }
 
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
     pk.pack(KEY_REQ_VALUES);
@@ -947,16 +984,19 @@ NetworkEngine::sendValueParts(Tid tid, const std::vector<Blob>& svals, const Soc
             end = std::min(start + MTU, v.size());
             buffer.clear();
             msgpack::packer<msgpack::sbuffer> pk(&buffer);
-            pk.pack_map(3+(config.network?1:0));
+            pk.pack_map(3+(config.network?1:0)+(config.is_client?1:0));
             if (config.network) {
                 pk.pack(KEY_NETID); pk.pack(config.network);
             }
+            if (config.is_client) {
+                pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+            }
             pk.pack(KEY_Y); pk.pack(KEY_V);
             pk.pack(KEY_TID); pk.pack(tid);
             pk.pack(KEY_V); pk.pack_map(1);
                 pk.pack(i); pk.pack_map(2);
-                    pk.pack(std::string("o")); pk.pack(start);
-                    pk.pack(std::string("d")); pk.pack_bin(end-start);
+                    pk.pack("o"sv); pk.pack(start);
+                    pk.pack("d"sv); pk.pack_bin(end-start);
                                                pk.pack_bin_body((const char*)v.data()+start, end-start);
             send(addr, buffer.data(), buffer.size());
             start = end;
@@ -971,7 +1011,7 @@ NetworkEngine::sendNodesValues(const SockAddr& addr, Tid tid, const Blob& nodes,
 {
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
-    pk.pack_map(4+(config.network?1:0));
+    pk.pack_map(4+(config.network?1:0)+(config.is_client?1:0));
 
     pk.pack(KEY_R);
     pk.pack_map(2 + (not st.empty()?1:0) + (nodes.size()>0?1:0) + (nodes6.size()>0?1:0) + (not token.empty()?1:0));
@@ -998,8 +1038,8 @@ NetworkEngine::sendNodesValues(const SockAddr& addr, Tid tid, const Blob& nodes,
             auto fields = query.select.getSelection();
             pk.pack(KEY_REQ_FIELDS);
             pk.pack_map(2);
-            pk.pack(std::string("f")); pk.pack(fields);
-            pk.pack(std::string("v")); pk.pack_array(st.size()*fields.size());
+            pk.pack("f"sv); pk.pack(fields);
+            pk.pack("v"sv); pk.pack_array(st.size()*fields.size());
             for (const auto& v : st)
                 v->msgpack_pack_fields(fields, pk);
             //DHT_LOG_DBG("sending closest nodes (%d+%d nodes.), %u value headers containing %u fields",
@@ -1009,10 +1049,13 @@ NetworkEngine::sendNodesValues(const SockAddr& addr, Tid tid, const Blob& nodes,
 
     pk.pack(KEY_TID); pk.pack(tid);
     pk.pack(KEY_Y); pk.pack(KEY_R);
-    pk.pack(KEY_UA); pk.pack(my_v);
+    pk.pack(KEY_UA); pk.pack(OPENDHT_UA);
     if (config.network) {
         pk.pack(KEY_NETID); pk.pack(config.network);
     }
+    if (config.is_client) {
+        pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+    }
 
     // send response
     send(addr, buffer.data(), buffer.size());
@@ -1073,7 +1116,7 @@ NetworkEngine::bufferNodes(sa_family_t af, const InfoHash& id, want_t want,
 }
 
 Sp<Request>
-NetworkEngine::sendListen(Sp<Node> n,
+NetworkEngine::sendListen(const Sp<Node>& n,
         const InfoHash& hash,
         const Query& query,
         const Blob& token,
@@ -1084,7 +1127,7 @@ NetworkEngine::sendListen(Sp<Node> n,
     Tid tid (n->getNewTid());
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
-    pk.pack_map(5+(config.network?1:0));
+    pk.pack_map(5+(config.network?1:0)+(config.is_client?1:0));
 
     auto has_query = not query.where.empty() or not query.select.empty();
     pk.pack(KEY_A); pk.pack_map(5 + has_query);
@@ -1100,10 +1143,13 @@ NetworkEngine::sendListen(Sp<Node> n,
     pk.pack(KEY_Q); pk.pack(QUERY_LISTEN);
     pk.pack(KEY_TID); pk.pack(tid);
     pk.pack(KEY_Y); pk.pack(KEY_Q);
-    pk.pack(KEY_UA); pk.pack(my_v);
+    pk.pack(KEY_UA); pk.pack(OPENDHT_UA);
     if (config.network) {
         pk.pack(KEY_NETID); pk.pack(config.network);
     }
+    if (config.is_client) {
+        pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+    }
 
     auto req = std::make_shared<Request>(MessageType::Listen, tid, n,
         Blob(buffer.data(), buffer.data() + buffer.size()),
@@ -1125,7 +1171,7 @@ void
 NetworkEngine::sendListenConfirmation(const SockAddr& addr, Tid tid) {
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
-    pk.pack_map(4+(config.network?1:0));
+    pk.pack_map(4+(config.network?1:0)+(config.is_client?1:0));
 
     pk.pack(KEY_R); pk.pack_map(2);
       pk.pack(KEY_REQ_ID); pk.pack(myid);
@@ -1133,16 +1179,19 @@ NetworkEngine::sendListenConfirmation(const SockAddr& addr, Tid tid) {
 
     pk.pack(KEY_TID); pk.pack(tid);
     pk.pack(KEY_Y); pk.pack(KEY_R);
-    pk.pack(KEY_UA); pk.pack(my_v);
+    pk.pack(KEY_UA); pk.pack(OPENDHT_UA);
     if (config.network) {
         pk.pack(KEY_NETID); pk.pack(config.network);
     }
+    if (config.is_client) {
+        pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+    }
 
     send(addr, buffer.data(), buffer.size());
 }
 
 Sp<Request>
-NetworkEngine::sendAnnounceValue(Sp<Node> n,
+NetworkEngine::sendAnnounceValue(const Sp<Node>& n,
         const InfoHash& infohash,
         const Sp<Value>& value,
         time_point created,
@@ -1153,13 +1202,14 @@ NetworkEngine::sendAnnounceValue(Sp<Node> n,
     Tid tid (n->getNewTid());
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
-    pk.pack_map(5+(config.network?1:0));
+    pk.pack_map(5+(config.network?1:0)+(config.is_client?1:0));
 
-    pk.pack(KEY_A); pk.pack_map((created < scheduler.time() ? 5 : 4));
+    bool add_created = created < scheduler.time();
+    pk.pack(KEY_A); pk.pack_map(add_created ? 5 : 4);
       pk.pack(KEY_REQ_ID);     pk.pack(myid);
       pk.pack(KEY_REQ_H);      pk.pack(infohash);
       auto v = packValueHeader(buffer, {value});
-      if (created < scheduler.time()) {
+      if (add_created) {
           pk.pack(KEY_REQ_CREATION);
           pk.pack(to_time_t(created));
       }
@@ -1168,10 +1218,13 @@ NetworkEngine::sendAnnounceValue(Sp<Node> n,
     pk.pack(KEY_Q);   pk.pack(QUERY_PUT);
     pk.pack(KEY_TID); pk.pack(tid);
     pk.pack(KEY_Y);   pk.pack(KEY_Q);
-    pk.pack(KEY_UA);  pk.pack(my_v);
+    pk.pack(KEY_UA);  pk.pack(OPENDHT_UA);
     if (config.network) {
         pk.pack(KEY_NETID); pk.pack(config.network);
     }
+    if (config.is_client) {
+        pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+    }
 
     auto req = std::make_shared<Request>(MessageType::AnnounceValue, tid, n,
         Blob(buffer.data(), buffer.data() + buffer.size()),
@@ -1199,27 +1252,53 @@ NetworkEngine::sendAnnounceValue(Sp<Node> n,
     return req;
 }
 
-Sp<Request>
-NetworkEngine::sendUpdateValues(Sp<Node> n,
+void
+NetworkEngine::sendUpdateValues(const Sp<Node>& n,
                                 const InfoHash& infohash,
-                                const std::vector<Sp<Value>>& values,
+                                std::vector<Sp<Value>>&& values,
                                 time_point created,
                                 const Blob& token,
-                                const size_t& socket_id)
+                                size_t socket_id)
+{
+    size_t total_size = 0;
+
+    auto b = values.begin(), e = values.begin();
+    while (e != values.end()) {
+        if (total_size >= MAX_MESSAGE_VALUE_SIZE) {
+            sendUpdateValues(n, infohash, b, e, created, token, socket_id);
+            b = e;
+            total_size = 0;
+        }
+        total_size += (*e)->size();
+        ++e;
+    }
+    if (b != e) {
+        sendUpdateValues(n, infohash, b, e, created, token, socket_id);
+    }
+}
+
+Sp<Request>
+NetworkEngine::sendUpdateValues(const Sp<Node>& n,
+                                const InfoHash& infohash,
+                                std::vector<Sp<Value>>::iterator begin,
+                                std::vector<Sp<Value>>::iterator end,
+                                time_point created,
+                                const Blob& token,
+                                size_t socket_id)
 {
     Tid tid (n->getNewTid());
     Tid sid (socket_id);
 
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
-    pk.pack_map(5+(config.network?1:0));
+    pk.pack_map(5+(config.network?1:0)+(config.is_client?1:0));
 
     pk.pack(KEY_A); pk.pack_map((created < scheduler.time() ? 7 : 6));
       pk.pack(KEY_REQ_ID);     pk.pack(myid);
       pk.pack(KEY_VERSION);    pk.pack(1);
       pk.pack(KEY_REQ_H);      pk.pack(infohash);
       pk.pack(KEY_REQ_SID);   pk.pack(sid);
-      auto v = packValueHeader(buffer, values);
+      auto v = packValueHeader(buffer, begin, end);
       if (created < scheduler.time()) {
           pk.pack(KEY_REQ_CREATION);
           pk.pack(to_time_t(created));
@@ -1229,10 +1308,13 @@ NetworkEngine::sendUpdateValues(Sp<Node> n,
     pk.pack(KEY_Q);   pk.pack(QUERY_UPDATE);
     pk.pack(KEY_TID); pk.pack(tid);
     pk.pack(KEY_Y);   pk.pack(KEY_Q);
-    pk.pack(KEY_UA);  pk.pack(my_v);
+    pk.pack(KEY_UA);  pk.pack(OPENDHT_UA);
     if (config.network) {
         pk.pack(KEY_NETID); pk.pack(config.network);
     }
+    if (config.is_client) {
+        pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+    }
 
     auto req = std::make_shared<Request>(MessageType::UpdateValue, tid, n,
         Blob(buffer.data(), buffer.data() + buffer.size()),
@@ -1246,7 +1328,7 @@ NetworkEngine::sendUpdateValues(Sp<Node> n,
 }
 
 Sp<Request>
-NetworkEngine::sendRefreshValue(Sp<Node> n,
+NetworkEngine::sendRefreshValue(const Sp<Node>& n,
                 const InfoHash& infohash,
                 const Value::Id& vid,
                 const Blob& token,
@@ -1257,7 +1339,7 @@ NetworkEngine::sendRefreshValue(Sp<Node> n,
     Tid tid (n->getNewTid());
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
-    pk.pack_map(5+(config.network?1:0));
+    pk.pack_map(5+(config.network?1:0)+(config.is_client?1:0));
 
     pk.pack(KEY_A); pk.pack_map(4);
       pk.pack(KEY_REQ_ID);       pk.pack(myid);
@@ -1268,10 +1350,13 @@ NetworkEngine::sendRefreshValue(Sp<Node> n,
     pk.pack(KEY_Q); pk.pack(QUERY_REFRESH);
     pk.pack(KEY_TID); pk.pack(tid);
     pk.pack(KEY_Y); pk.pack(KEY_Q);
-    pk.pack(KEY_UA); pk.pack(my_v);
+    pk.pack(KEY_UA); pk.pack(OPENDHT_UA);
     if (config.network) {
         pk.pack(KEY_NETID); pk.pack(config.network);
     }
+    if (config.is_client) {
+        pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+    }
 
     auto req = std::make_shared<Request>(MessageType::Refresh, tid, n,
         Blob(buffer.data(), buffer.data() + buffer.size()),
@@ -1303,7 +1388,7 @@ void
 NetworkEngine::sendValueAnnounced(const SockAddr& addr, Tid tid, Value::Id vid) {
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
-    pk.pack_map(4+(config.network?1:0));
+    pk.pack_map(4+(config.network?1:0)+(config.is_client?1:0));
 
     pk.pack(KEY_R); pk.pack_map(3);
       pk.pack(KEY_REQ_ID);  pk.pack(myid);
@@ -1312,10 +1397,13 @@ NetworkEngine::sendValueAnnounced(const SockAddr& addr, Tid tid, Value::Id vid)
 
     pk.pack(KEY_TID); pk.pack(tid);
     pk.pack(KEY_Y); pk.pack(KEY_R);
-    pk.pack(KEY_UA); pk.pack(my_v);
+    pk.pack(KEY_UA); pk.pack(OPENDHT_UA);
     if (config.network) {
         pk.pack(KEY_NETID); pk.pack(config.network);
     }
+    if (config.is_client) {
+        pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+    }
 
     send(addr, buffer.data(), buffer.size());
 }
@@ -1329,7 +1417,7 @@ NetworkEngine::sendError(const SockAddr& addr,
 {
     msgpack::sbuffer buffer;
     msgpack::packer<msgpack::sbuffer> pk(&buffer);
-    pk.pack_map(4 + (include_id?1:0) + (config.network?1:0));
+    pk.pack_map(4 + (include_id?1:0) + (config.network?1:0)+(config.is_client?1:0));
 
     pk.pack(KEY_E); pk.pack_array(2);
       pk.pack(code);
@@ -1342,10 +1430,13 @@ NetworkEngine::sendError(const SockAddr& addr,
 
     pk.pack(KEY_TID); pk.pack(tid);
     pk.pack(KEY_Y); pk.pack(KEY_E);
-    pk.pack(KEY_UA); pk.pack(my_v);
+    pk.pack(KEY_UA); pk.pack(OPENDHT_UA);
     if (config.network) {
         pk.pack(KEY_NETID); pk.pack(config.network);
     }
+    if (config.is_client) {
+        pk.pack(KEY_ISCLIENT); pk.pack(config.is_client);
+    }
 
     send(addr, buffer.data(), buffer.size());
 }
